"""Seeking Alpha Forward Sales Estimates Model."""

# pylint: disable=unused-argument

from typing import Any, Dict, List, Literal, Optional
from warnings import warn

from openbb_core.app.model.abstract.error import OpenBBError
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.forward_sales_estimates import (
    ForwardSalesEstimatesData,
    ForwardSalesEstimatesQueryParams,
)
from openbb_core.provider.utils.helpers import amake_request
from openbb_seeking_alpha.utils.helpers import HEADERS, get_seekingalpha_id
from pydantic import Field, field_validator


class SAForwardSalesEstimatesQueryParams(ForwardSalesEstimatesQueryParams):
    """Seeking Alpha Forward Sales Estimates Query.

    Source: https://seekingalpha.com/earnings/earnings-calendar
    """

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}

    period: Literal["annual", "quarter"] = Field(
        default="quarter",
        description="The reporting period.",
        json_schema_extra={"choices": ["annual", "quarter"]},
    )

    @field_validator("symbol", mode="before", check_fields=False)
    @classmethod
    def check_symbol(cls, value):
        """Check the symbol."""
        if not value:
            raise OpenBBError("Symbol is a required field for Seeking Alpha.")
        return value


class SAForwardSalesEstimatesData(ForwardSalesEstimatesData):
    """Seeking Alpha Forward Sales Estimates Data."""

    actual: Optional[int] = Field(
        default=None,
        description="Actual sales (revenue) for the period.",
    )
    period_growth: Optional[float] = Field(
        default=None,
        description="Estimated (or actual if reported) EPS growth for the period.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )


class SAForwardSalesEstimatesFetcher(
    Fetcher[
        SAForwardSalesEstimatesQueryParams,
        List[SAForwardSalesEstimatesData],
    ]
):
    """Seeking Alpha Forward Sales Estimates Fetcher."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> SAForwardSalesEstimatesQueryParams:
        """Transform the query."""
        return SAForwardSalesEstimatesQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: SAForwardSalesEstimatesQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> Dict:
        """Return the raw data from the Seeking Alpha endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_core.provider.utils.client import ClientSession

        tickers = query.symbol.split(",")  # type: ignore
        fp = query.period if query.period == "annual" else "quarterly"
        url = "https://seekingalpha.com/api/v3/symbol_data/estimates"
        querystring = {
            "estimates_data_items": "revenue_actual,revenue_consensus_low,revenue_consensus_mean,"
            "revenue_consensus_high,revenue_num_of_estimates",
            "period_type": fp,
            "relative_periods": "-4,-3,-2,-1,0,1,2,3,4,5,6,7,8,9,10,11,12",
        }

        async with ClientSession(trust_env=True) as session:
            kwargs = {"session": session}
            ids = {
                ticker: await get_seekingalpha_id(ticker, **kwargs)
                for ticker in tickers
            }
            querystring["ticker_ids"] = ("%2C").join(list(ids.values()))
            response = await amake_request(
                url, headers=HEADERS, params=querystring, session=session
            )
            estimates = response.get("estimates", {})  # type: ignore

        if not estimates:
            raise OpenBBError(f"No estimates data was returned for: {query.symbol}")
        output: Dict = {"ids": ids, "estimates": estimates}

        return output

    @staticmethod
    def transform_data(
        query: SAForwardSalesEstimatesQueryParams,
        data: Dict,
        **kwargs: Any,
    ) -> List[SAForwardSalesEstimatesData]:
        """Transform the data to the standard format."""
        tickers = query.symbol.split(",")  # type: ignore
        ids = data.get("ids", {})
        estimates = data.get("estimates", {})
        results: List[SAForwardSalesEstimatesData] = []
        for ticker in tickers:
            sa_id = str(ids.get(ticker, ""))
            if sa_id == "" or sa_id not in estimates:
                warn(f"Symbol Error: No data found for, {ticker}")
            seek_object = estimates.get(sa_id, {})
            if not seek_object:
                warn(f"No data found for {ticker}")
                continue
            items = len(seek_object.get("revenue_num_of_estimates"))
            if not items:
                warn(f"No data found for {ticker}")
                continue
            for i in range(0, items - 4):
                rev_estimates: Dict = {}
                rev_estimates["symbol"] = ticker
                num_estimates = seek_object["revenue_num_of_estimates"].get(str(i))
                if not num_estimates:
                    continue
                period = num_estimates[0].get("period", {})
                if period:
                    period_type = period.get("periodtypeid")
                    rev_estimates["calendar_year"] = period.get("calendaryear")
                    rev_estimates["calendar_period"] = (
                        "Q" + str(period.get("calendarquarter", ""))
                        if period_type == "quarterly"
                        else "FY"
                    )
                    rev_estimates["date"] = period.get("periodenddate").split("T")[0]
                    rev_estimates["fiscal_year"] = period.get("fiscalyear")
                    rev_estimates["fiscal_period"] = (
                        "Q" + str(period.get("fiscalquarter", ""))
                        if period_type == "quarterly"
                        else "FY"
                    )
                rev_estimates["number_of_analysts"] = num_estimates[0].get(
                    "dataitemvalue"
                )
                mean = seek_object["revenue_consensus_mean"].get(str(i))
                if mean:
                    mean = mean[0].get("dataitemvalue")
                    rev_estimates["mean"] = int(float(mean))
                actual = (
                    seek_object["revenue_actual"][str(i)][0].get("dataitemvalue")
                    if i < 1
                    else None
                )
                if actual:
                    rev_estimates["actual"] = int(float(actual))
                low = seek_object["revenue_consensus_low"].get(str(i))
                if low:
                    low = low[0].get("dataitemvalue")
                    rev_estimates["low_estimate"] = int(float(low))
                high = seek_object["revenue_consensus_high"].get(str(i))
                if high:
                    high = high[0].get("dataitemvalue")
                    rev_estimates["high_estimate"] = int(float(high))
                # Calculate the estimated growth percent.
                this = float(mean) if mean else None
                prev = None
                percent = None
                try:
                    prev = float(
                        seek_object["revenue"][str(i - 1)][0].get("dataitemvalue")
                    )
                except KeyError:
                    prev = float(
                        seek_object["revenue_consensus_mean"][str(i - 1)][0].get(
                            "dataitemvalue"
                        )
                    )
                if this and prev:
                    percent = (this - prev) / prev
                rev_estimates["period_growth"] = percent
                results.append(
                    SAForwardSalesEstimatesData.model_validate(rev_estimates)
                )

        return results



"""Seeking Alpha Utilities."""

from datetime import timedelta

from openbb_core.provider.utils.helpers import amake_request

HEADERS = {
    "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/128.0.0.0",  # noqa
    "Accept": "*/*",
    "Accept-Language": "en-US,en;q=0.9",
    "Accept-Encoding": "gzip, deflate, br, zstd",
    "Sec-Fetch-Dest": "empty",
    "Sec-Fetch-Mode": "cors",
    "Sec-Fetch-Site": "same-origin",
    "Connection": "keep-alive",
}


def date_range(start_date, end_date):
    """Yield dates between start_date and end_date."""
    for n in range(int((end_date - start_date).days) + 1):
        yield start_date + timedelta(n)


async def get_seekingalpha_id(symbol: str, **kwargs) -> str:
    """Map a ticker symbol to its Seeking Alpha ID."""
    session = kwargs.get("session")
    url = "https://seekingalpha.com/api/v3/searches"
    querystring = {
        "filter[type]": "symbols",
        "filter[list]": "all",
        "page[size]": "100",
    }
    querystring["filter[query]"] = symbol
    if session:
        response = await amake_request(
            url, headers=HEADERS, params=querystring, session=session
        )
    else:
        response = await amake_request(url, headers=HEADERS, params=querystring)
    ids = response.get("symbols")  # type: ignore

    return str(ids[0].get("id", "")) if isinstance(ids, list) and ids else ""



"""Seeking Alpha Provider module."""

from openbb_core.provider.abstract.provider import Provider
from openbb_seeking_alpha.models.calendar_earnings import SACalendarEarningsFetcher
from openbb_seeking_alpha.models.forward_eps_estimates import (
    SAForwardEpsEstimatesFetcher,
)
from openbb_seeking_alpha.models.forward_sales_estimates import (
    SAForwardSalesEstimatesFetcher,
)

seeking_alpha_provider = Provider(
    name="seeking_alpha",
    website="https://seekingalpha.com",
    description="""Seeking Alpha is a data provider with access to news, analysis, and
real-time alerts on stocks.""",
    fetcher_dict={
        "CalendarEarnings": SACalendarEarningsFetcher,
        "ForwardEpsEstimates": SAForwardEpsEstimatesFetcher,
        "ForwardSalesEstimates": SAForwardSalesEstimatesFetcher,
    },
    repr_name="Seeking Alpha",
)



"""Seeking Alpha Calendar Earnings Model."""

# pylint: disable=unused-argument

import asyncio
import json
from datetime import datetime, timedelta
from typing import Any, Dict, List, Literal, Optional
from warnings import warn

from openbb_core.app.model.abstract.error import OpenBBError
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.calendar_earnings import (
    CalendarEarningsData,
    CalendarEarningsQueryParams,
)
from openbb_seeking_alpha.utils.helpers import HEADERS, date_range
from pydantic import Field, field_validator


class SACalendarEarningsQueryParams(CalendarEarningsQueryParams):
    """Seeking Alpha Calendar Earnings Query.

    Source: https://seekingalpha.com/earnings/earnings-calendar
    """

    country: Literal["us", "ca"] = Field(
        default="us",
        description="The country to get calendar data for.",
        json_schema_extra={"choices": ["us", "ca"]},
    )


class SACalendarEarningsData(CalendarEarningsData):
    """Seeking Alpha Calendar Earnings Data."""

    market_cap: Optional[float] = Field(
        default=None,
        description="Market cap of the entity.",
    )
    reporting_time: Optional[str] = Field(
        default=None,
        description="The reporting time - e.g. after market close.",
    )
    exchange: Optional[str] = Field(
        default=None,
        description="The primary trading exchange.",
    )
    sector_id: Optional[int] = Field(
        default=None,
        description="The Seeking Alpha Sector ID.",
    )

    @field_validator("report_date", mode="before", check_fields=False)
    @classmethod
    def validate_release_date(cls, v):
        """Validate the release date."""
        v = v.split("T")[0]
        return datetime.strptime(v, "%Y-%m-%d").date()


class SACalendarEarningsFetcher(
    Fetcher[
        SACalendarEarningsQueryParams,
        List[SACalendarEarningsData],
    ]
):
    """Seeking Alpha Calendar Earnings Fetcher."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> SACalendarEarningsQueryParams:
        """Transform the query."""
        now = datetime.today().date()
        transformed_params = params
        if not params.get("start_date"):
            transformed_params["start_date"] = now
        if not params.get("end_date"):
            transformed_params["end_date"] = now + timedelta(days=3)
        return SACalendarEarningsQueryParams(**transformed_params)

    @staticmethod
    async def aextract_data(
        query: SACalendarEarningsQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the Seeking Alpha endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_core.provider.utils.client import ClientSession
        from openbb_core.provider.utils.helpers import amake_request

        results: List[Dict] = []
        dates = [
            date.strftime("%Y-%m-%d")
            for date in date_range(query.start_date, query.end_date)
        ]
        currency = "USD" if query.country == "us" else "CAD"
        messages: List = []

        async with ClientSession() as session:

            async def get_date(date, currency):
                """Get date for one date."""
                url = (
                    f"https://seekingalpha.com/api/v3/earnings_calendar/tickers?"
                    f"filter%5Bselected_date%5D={date}"
                    f"&filter%5Bwith_rating%5D=false&filter%5Bcurrency%5D={currency}"
                )
                response = await amake_request(
                    url=url, headers=HEADERS, session=session
                )
                # Try again if the response is blocked.
                if "blockScript" in response:
                    response = await amake_request(
                        url=url, headers=HEADERS, session=session
                    )
                    if "blockScript" in response:
                        message = json.dumps(response)
                        messages.append(message)
                        warn(message)
                if "data" in response:
                    results.extend(response.get("data"))  # type: ignore

            await asyncio.gather(*[get_date(date, currency) for date in dates])

        if not results:
            raise OpenBBError(f"Error with the Seeking Alpha request -> {messages}")

        return results

    @staticmethod
    def transform_data(
        query: SACalendarEarningsQueryParams,
        data: List[Dict],
        **kwargs: Any,
    ) -> List[SACalendarEarningsData]:
        """Transform the data to the standard format."""
        transformed_data: List[SACalendarEarningsData] = []
        for row in sorted(data, key=lambda x: x["attributes"]["release_date"]):
            attributes = row.get("attributes", {})
            transformed_data.append(
                SACalendarEarningsData.model_validate(
                    {
                        "report_date": attributes.get("release_date"),
                        "reporting_time": attributes.get("release_time"),
                        "symbol": attributes.get("slug"),
                        "name": attributes.get("name"),
                        "market_cap": attributes.get("marketcap"),
                        "exchange": attributes.get("exchange"),
                        "sector_id": attributes.get("sector_id"),
                    }
                )
            )
        return transformed_data
